package com.FranckBarbier.Java._Pedagogical_RSA;

public class Euclid {

    private long _a, _b, _gcd, _x, _y;

    public Euclid(long a, long b) {
        if (a < 0L) {
            _a = -a;
        } else {
            _a = a;
        }
        if (b < 0L) {
            _b = -b;
        } else {
            _b = b;
        }
        if (_a == 0L) {
            _gcd = _b;
        } else if (_b == 0L) {
            _gcd = _a;
        } else {
            _gcd = extended_Euclid(_a, _b);
        }
    }

    private long extended_Euclid(long a, long b) {
        if (b == 0L) {
            _x = 1L;
            _y = 0L;
            return a;
        }
        _gcd = extended_Euclid(b, a % b);
        long xp = _y;
        long yp = _x - (a / b) * _y;
        _x = xp;
        _y = yp;
        return _gcd;
    }

    public long a() {
        return _a;
    }

    public long b() {
        return _b;
    }

    public long gcd() {
        return _gcd;
    }

    public long x() {
        return _x;
    }

    public long y() {
        return _y;
    }

    public static long Basic_Euclid(long a, long b) {
        if (a < 0L) {
            a = -a;
        }
        if (b < 0L) {
            b = -b;
        }
        if (a == 0L) {
            return b;
        } else if (b == 0L) {
            return a;
        } else {
            return Basic_Euclid(b, a % b);
        }
    }

    public long[] solver(long b) throws Exception { // a * x (mod n) = b
        long n = b(); // we have gcd = a * x + n * y, so requiring b() leads to have n
        if (Basic_Euclid(gcd(), b) == gcd()) { // gcd(a,n) divides b? <=> gcd(gcd(a,n),b) = gcd(a,n)?
            long d[] = new long[(int) gcd()]; // Risk of error here since gcd() returns a 'long' but no worries...
            d[0] = (x() * (b / gcd())) % n;
            if (d[0] < 0) {
                d[0] = n + d[0]; // For instance [-5]100 = [95]100
            }
            for (int i = 1; i < d.length; i++) {
                d[i] = d[0] + (i * (n / gcd())) % n;
            }
            return d;
        }
        throw new Exception("Euclid: No solution");
    }

    /**
     * x (i.e., e or d) is transformed into its binary representation for
     * exponentiation
     */
    public static java.util.BitSet toBitSet(long e) {
        java.util.BitSet bs = new java.util.BitSet();
        long powersOfTwo = 0x0000000000000001L;
        for (int i = 0; i < 64; i++) {
            if ((e & powersOfTwo) == powersOfTwo) {
                bs.set(i);
            }
            powersOfTwo *= 0x0000000000000002L;
        }
        return bs;
    }

    /**
     * e times of message * message *... (mod n) or d times of cipher message *
     * cipher message *... (mod n)
     */
    public static long Exponentiation(long message, long n, java.util.BitSet e) {
        if (e.length() == 0) {
            return 1L;
        }
        long cipher_message = 1L;
        for (int i = e.length(); i > 0; i--) {
            cipher_message = (cipher_message * cipher_message) % n;
            if (e.get(i - 1)) {
                cipher_message = (cipher_message * message) % n;
            }
        }
        return cipher_message;
    }
}
